module stream;
// not copyrighted
// written by Dario Schiavon (supdar@yahoo.com)

// I took inspiration from Pavel Minayev's stream.d (c)
// and tried to write a new stream.d.
// My primary intent was to provide a set of functions
// for formatted input and output (see print() and scan()).

class Stream
{
	// Base class for the streams.
	// This class is not meant to be used:
	// use File, Socket, MemoryStream or SliceStream instead.

	abstract void writeBlock(void*, uint);	// THESE SHOULD BE PROTECTED BUT
	abstract void readBlock(void*, uint);	// stream_socket.d COMPLAINS

	// Use these functions to write raw data into
	// a stream. (They are inherited by all the streams.)

	void write(in char a)	{ writeBlock(&a, a.size); }
	void write(in wchar a)	{ writeBlock(&a, a.size); }

	void write(in byte a)	{ writeBlock(&a, a.size); }
	void write(in ubyte a)	{ writeBlock(&a, a.size); }
	void write(in short a)	{ writeBlock(&a, a.size); }
	void write(in ushort a)	{ writeBlock(&a, a.size); }
	void write(in int a)	{ writeBlock(&a, a.size); }
	void write(in uint a)	{ writeBlock(&a, a.size); }
	void write(in long a)	{ writeBlock(&a, a.size); }
	void write(in ulong a)	{ writeBlock(&a, a.size); }

	void write(in float a)	{ writeBlock(&a, a.size); }
	void write(in double a)	{ writeBlock(&a, a.size); }
	void write(in real a)	{ writeBlock(&a, a.size); }

	void write(in ifloat a)	{ writeBlock(&a, a.size); }
	void write(in idouble a){ writeBlock(&a, a.size); }
	void write(in ireal a)	{ writeBlock(&a, a.size); }

	void write(in cfloat a)	{ writeBlock(&a, a.size); }
	void write(in cdouble a){ writeBlock(&a, a.size); }
	void write(in creal a)	{ writeBlock(&a, a.size); }

	void write(in char[] a)	{ writeBlock(a, a.length * (*a).size); }
	void write(in wchar[] a){ writeBlock(a, a.length * (*a).size); }
	void write(in byte[] a)	{ writeBlock(a, a.length * (*a).size); }
	void write(in ubyte[] a){ writeBlock(a, a.length * (*a).size); }

	// Use these functions to write the array length
	// followed by the arrays itself. (Then use
	// readStringWithLength(), readWstringWithLength(),
	// readByteArrayWithLength() and readUbyteArrayWithLength()
	// to retrieve it.)

	void writeArrayWithLength(in char[] a)	{ write(a.length); write(cast(ubyte[]) a); }
	void writeArrayWithLength(in wchar[] a)	{ write(a.length); write(cast(ubyte[]) a); }
	void writeArrayWithLength(in byte[] a)	{ write(a.length); write(cast(ubyte[]) a); }
	void writeArrayWithLength(in ubyte[] a)	{ write(a.length); write(cast(ubyte[]) a); }

	// Use these functions to read raw data from
	// a stream. You can also use readInt(), readChar(), etc.

	void read(out byte a)	{ readBlock(&a, a.size); }
	void read(out ubyte a)	{ readBlock(&a, a.size); }
	void read(out short a)	{ readBlock(&a, a.size); }
	void read(out ushort a)	{ readBlock(&a, a.size); }
	void read(out int a)	{ readBlock(&a, a.size); }
	void read(out uint a)	{ readBlock(&a, a.size); }
	void read(out long a)	{ readBlock(&a, a.size); }
	void read(out ulong a)	{ readBlock(&a, a.size); }

	void read(out float a)	{ readBlock(&a, a.size); }
	void read(out double a)	{ readBlock(&a, a.size); }
	void read(out real a)	{ readBlock(&a, a.size); }

	void read(out ifloat a)	{ readBlock(&a, a.size); }
	void read(out idouble a){ readBlock(&a, a.size); }
	void read(out ireal a)	{ readBlock(&a, a.size); }

	void read(out cfloat a)	{ readBlock(&a, a.size); }
	void read(out cdouble a){ readBlock(&a, a.size); }
	void read(out creal a)	{ readBlock(&a, a.size); }

	void read(out char a)	{ readBlock(&a, a.size); }
	void read(out wchar a)	{ readBlock(&a, a.size); }

	void read(in char[] a)	{ readBlock(a, a.length * (*a).size); } // these functions will
	void read(in wchar[] a)	{ readBlock(a, a.length * (*a).size); } // read as much data as
	void read(in byte[] a)	{ readBlock(a, a.length * (*a).size); } // the array is large.
	void read(in ubyte[] a)	{ readBlock(a, a.length * (*a).size); }

	// You will probably find these functions
	// simpler to use than read(out...).   =)

	byte    readByte()	{ byte a;    read(a); return a; }
	ubyte   readUbyte()	{ ubyte a;   read(a); return a; }
	short   readShort()	{ short a;   read(a); return a; }
	ushort  readUshort()	{ ushort a;  read(a); return a; }
	int     readInt()	{ int a;     read(a); return a; }
	uint    readUint()	{ uint a;    read(a); return a; }
	long    readLong()	{ long a;    read(a); return a; }
	ulong   readUlong()	{ ulong a;   read(a); return a; }

	float   readFloat()	{ float a;   read(a); return a; }
	double  readDouble()	{ double a;  read(a); return a; }
	real    readReal()	{ real a;    read(a); return a; }

	ifloat  readIfloat()	{ ifloat a;  read(a); return a; }
	idouble readIdouble()	{ idouble a; read(a); return a; }
	ireal   readIreal()	{ ireal a;   read(a); return a; }

	cfloat  readCfloat()	{ cfloat a;  read(a); return a; }
	cdouble readCdouble()	{ cdouble a; read(a); return a; }
	creal   readCreal()	{ creal a;   read(a); return a; }

	char    readChar()	{ char a;    read(a); return a; }
	wchar   readWchar()	{ wchar a;   read(a); return a; }

	// These functions allocate the arrays by themselves.

	char[]  readString(uint len)	{ char[] s   = new char[len];  read(s);  return s;  }
	wchar[] readWstring(uint len)	{ wchar[] ws = new wchar[len]; read(ws); return ws; }
	byte[]  readByteArray(uint len)	{ byte[] ba  = new byte[len];  read(ba); return ba; }
	ubyte[] readUbyteArray(uint len){ ubyte[] ua = new ubyte[len]; read(ua); return ua; }

	// These functions read an array which has been
	// written by writeArrayWithLength().

	char[]	readStringWithLength()	  { return readString(readUint());     }
	wchar[]	readWstringWithLength()	  { return readWstring(readUint());    }
	byte[]	readByteArrayWithLength() { return readByteArray(readUint());  }
	ubyte[]	readUbyteArrayWithLength(){ return readUbyteArray(readUint()); }

	// These functions will write formatted data
	// into a stream.
	// Numbers are converted into their decimal
	// representation. A period is used as to
	// separate characteristic from mantissa.
	// It is necessary to specify how much precision
	// floating point numbers will be printed with.
	// If no precision is specified only their
	// characteristic is printed.
	// Note that char[] is used, not wchar[].

	void print(in byte a)	{ print(cast(int) a);  }
	void print(in ubyte a)	{ print(cast(uint) a); }
	void print(in short a)	{ print(cast(int) a);  }
	void print(in ushort a)	{ print(cast(uint) a); }
	void print(in int a)
	{
		if(a < 0)
		{	write(cast(char) '-');
			a = -a;
		}
		print(cast(uint) a);
	}
	void print(in uint a)
	{
		if(a == 0) write(cast(char) '0');
		char[10] s;
		uint i = s.length - 1;
		while(a)
		{	s[i] = cast(char) (a % 10) + '0';
			a /= 10;
			i--;
		}
		write(s[i+1 .. s.length]);
	}
	void print(in long a)
	{
		if(a < 0)
		{	write(cast(char) '-');
			a = -a;
		}
		print(cast(ulong) a);
	}
	void print(in ulong a)
	{
		if(a == 0) write(cast(char) '0');
		char[20] s;
		uint i = s.length - 1;
		while(a)
		{	s[i] = cast(char) (a % 10) + '0';
			a /= 10;
			i--;
		}
		write(s[i+1 .. s.length]);
	}

	void print(in float a)	{ print(cast(long) a); } // TO BE FIXED:
	void print(in double a)	{ print(cast(long) a); } // REALLY GREAT NUMBERS WON'T
	void print(in real a)	{ print(cast(long) a); } // BE PRINTED CORRECTLY!

	import math;
	void print(in float a, int p)	{ print(cast(double) a, p); }
	void print(in real a, int p)	{ print(cast(double) a, p); }
	void print(in double a, int p)	// TO BE FIXED!
	{
		long c = a;
		ulong m = fabs((a - cast(real) c) * pow(cast(real) 10, p));

		char[20] s;
		uint i = s.length - 1;
		while(i >= s.length - p)
		{
			s[i] = cast(char) (m % 10) + '0';
			m /= 10;
			i--;
		}

		print(c);
		write(cast(char) '.');
		write(s[i+1 .. s.length]);
	}

	void print(in char[] s)
	{
	    version(Win32)
	    {
		static char[2] newl = \r\n;
		for(uint i=0; i<s.length; i++)
			if(s[i] == \n) write(newl);
			else write(s[i]);
	    }
	    version(linux)
	    {
		write(s);
	    }
	}
	void print(in wchar[] s)
	{
	    version(Win32)
	    {
		static wchar[2] newl = \r\n;
		for(uint i=0; i<s.length; i++)
			if(s[i] == \n) write(newl);
			else write(s[i]);
	    }
	    version(linux)
	    {
		write(s);
	    }
	}

	/++++++++++++++++++++ PROBLEMS LINKING conv.d
	// These functions will read formatted data
	// from a stream and convert them into their
	// binary representation.

	void scan(out byte a)	{ a = scanByte();   }
	void scan(out ubyte a)	{ a = scanUbyte();  }
	void scan(out short a)	{ a = scanShort();  }
	void scan(out ushort a)	{ a = scanUshort(); }
	void scan(out int a)	{ a = scanInt();    }
	void scan(out uint a)	{ a = scanUint();   }
	void scan(out long a)	{ a = scanLong();   }
	void scan(out ulong a)	{ a = scanUlong();  }

	import ctype, conv;
	byte	scanByte()	{ return scanInt();  }
	ubyte	scanUbyte()	{ return scanUint(); }
	short	scanShort()	{ return scanInt();  }
	ushort	scanUshort()	{ return scanUint(); }
	int	scanInt()
	{
		char[10] s;
		uint i = 0;

		read(s[0]);
		if(s[0] == '+') return +scanUint();
		if(s[0] == '-') return -scanUint();

		while(isdigit(s[i++]))
			read(s[i]);

		return conv.toUint(s[0 .. i]);
	}
	uint	scanUint()
	{
		char[10] s;
		uint i = 0;

		while(read(s[i]), isdigit(s[i++]))
			continue;

		return conv.toUint(s[0 .. i]);
	}
	long	scanLong()
	{
		char[20] s;
		uint i = 0;

		read(s[0]);
		if(s[0] == '+') return +scanUlong();
		if(s[0] == '-') return -scanUlong();

		while(isdigit(s[i++]))
			read(s[i]);

		return conv.toUlong(s[0 .. i]);
	}
	ulong	scanUlong()
	{
		char[20] s;
		uint i = 0;

		while(read(s[i]), isdigit(s[i++]))
			continue;

		return conv.toUlong(s[0 .. i]);
	}

	// WHY DOESN'T float conv.toFloat(char[]) EXIST?
	// WHY DOESN'T double conv.toDouble(char[]) EXIST?
	// WHY DOESN'T real conv.toReal(char[]) EXIST?
	++++++++++++++++++++/

	// Use get() to get the entire array (as an ubyte[]).
	// (get() will not move the stream pointer.)

	ubyte[] get()
	{
		ulong pos = position();
		seek(0, Rel.begin);
		ubyte[] arr = readUbyteArray(length());
		seek(pos, Rel.begin);
		return arr;
	}

	// Use seek() to move the stream pointer.
	// Use position() to get the position of the stream pointer.
	// Use length() to read the stream length.
	// Use truncate() to truncate a stream.

	abstract void seek(int, Rel);
	abstract ulong position();
	abstract ulong length();
	abstract void truncate();
}

enum Rel : uint { begin = 0, cur = 1, end = 2 }

class StreamError : Exception
{
	// Base class for stream exceptions,
	// thrown when the error doesn't suit
	// any other exception class.

	this(char[] msg)
	{ super(msg); }
}

class ReadError : StreamError
{
	// Thrown when unable to read data.

	this(char[] msg)
	{ super(msg); }
}

class WriteError : StreamError
{
	// Thrown when unable to write data.

	this(char[] msg)
	{ super(msg); }
}

class SeekError : StreamError
{
	// Thrown when unable to move stream pointer.

	this(char[] msg)
	{ super(msg); }
}

public
{
	import stream_memory;
	import stream_file;
	import stream_socket;
	import stream_slice;
}