module stream_memory;
import stream;

class MemoryStream : Stream
{
	// This stream is bound to an array in memory.
	// You can pass an array to the constructor
	// so that it will use that array (until a
	// reallocation is needed).

	this()		{ this(250); }
	this(uint sz)
	{
		data.length = len = sz;
		clear(); // IS THIS NEEDED?
	}

	this(byte[] a)	{ this(cast(ubyte[]) a); }
	this(ubyte[] a)
	in
	{
		assert(a);
	}
	body
	{
		data = a;
		len = a.length;
	}

	this(char[] s)	{ this(cast(ubyte[]) s); }
	this(wchar[] s)
	{
		this((cast(ubyte*)cast(wchar*) s)[0 .. s.length * wchar.size]);
	}

	// When finished using the stream
	// call get() to get the array.

	ubyte[] get()
	{
		return data[0 .. len];
	}

	// Call clear() to zero the stream's array.

	void clear()
	{
		ubyte* end = cast(ubyte*) data + len;
		for(ubyte* i = data; i < end; i++)
			*i = 0x00;
	}

	// Call reserve() to ensure that the writing
	// of a certain amount of data will not cause
	// a reallocation.

	void reserve(uint amount)
	{
		uint reqlen = i + amount;
		if(reqlen > len)
			if(reqlen > data.length) data.length = reqlen;
			else len = reqlen;
	}

	void seek(int offset, Rel rel)
	in
	{
		if(rel == Rel.begin) assert(offset >= 0);
		if(rel == Rel.end)   assert(offset <= 0);
	}
	body
	{
		switch(rel)
		{
			case Rel.begin:	i = offset; break;
			case Rel.cur:	i += offset; break;
			case Rel.end:	i = len + offset; break;
		}
		if(i > len) // i is uint, so cannot be < 0
			throw new SeekError("MemoryStream: pointer outside of stream bounds");
	}

	void truncate()
	{
		len = i;
	}

	ulong length()
	{
		return len;
	}

		protected:

	ubyte[] data;
	uint i = 0, len;

	void readBlock(void* buf, uint buflen)
	{
		assert(buf);
		if(i + buflen > len)
			throw new ReadError("MemoryStream: no enough data in stream");

		uint cur = i;
		i += buflen;
		(cast(ubyte*) buf)[0 .. buflen] = data[cur .. i];
	}

	void writeBlock(void* buf, uint buflen)
	{
		assert(buf);
		uint cur = i;
		i += buflen;

		if(i > len) // i = required length
			if(i > data.length) data.length = i * 2; // reserve extra space
			else len = i;

		data[cur .. i] = (cast(ubyte*) buf)[0 .. buflen];
	}
}