module stream_socket;
import stream;

version(Win32) private
{
	// from WINSOCK.H
	struct WSAData
	{
		ushort wVersion;
		ushort wHighVersion;
		char[257] szDescription;
		char[129] szSystemStatus;
		ushort iMaxSockets;
		ushort iMaxUdpDg;
		char* lpVendorInfo;
	}

	struct in_addr
	{
		uint s_addr;
	}

	struct sockaddr_in
	{
		short sin_family;
		ushort sin_port;
		in_addr sin_addr;
		byte[8] sin_zero;
	}

	extern(Windows):

	int WSAStartup(ushort, WSAData*);
	int WSACleanup();

	int socket(int, int, int);
	int closesocket(int);

	int connect(int, sockaddr_in*, uint);
	int bind(int, sockaddr_in*, uint);
	int listen(int , int);
	int accept(int, sockaddr_in*, uint*);

	int send(int, char*, int, int);
	int recv(int, char*, int, int);
}
version(linux)
{
	// WHAT SHOULD BE IMPORTED?
}

alias SocketStream Socket;
class SocketStream: Stream
{
	// This stream is bound to a socket,
	// so every time it is written a packet
	// is sent.
	// [THIS STREAM HASN'T AN OUTPUT BUFFER,
	// MAYBE IT SHOULD BE ADDED.]

	// Two SocketStreams are needed: the
	// former listens for connections, the
	// latter connects to the former.

	// If the constructor is passed an ip
	// address it is supposed to establish
	// a connection, otherwise it will wait
	// for a connection.

	// Both the sockets must share the same
	// port. (Actually these sockets are based
	// on the TCP/IP protocol, other protocols
	// can be implemented anyway.)

	// Ip addresses can be passed both as
	// strings and as uints. Uints must be
	// big-endian (network byte order).

	import ctype;
	this(char[] s, ushort port)
	{
		uint addr = 0;
		ubyte octet = 0;
		uint i = -1;

		while(++i != s.length)
		if(s[i] == '.')
		{
			addr >>= 8;
			addr += octet << 24;
			octet = 0;
		}
		else if(isdigit(s[i]))
		{
			octet *= 10;
			octet += s[i] - '0';
		}
		else throw new StreamError("SocketStream: invalid ip address");

		this((addr >> 8) + (octet << 24), port);
	}
	this(uint address, ushort port)
	{
		connectTo(address, port);
	}

	this(ushort port)
	{
		listenTo(port);
	}

	~this()
	{
		if(skt != -1) disconnect();
	}

	void connectTo(uint address, ushort port)
	{
		sockaddr_in addr;
		addr.sin_family = 2;
		addr.sin_port = port;
		addr.sin_addr.s_addr = address;

		skt = socket(addr.sin_family, 1, 0);
		if(skt == -1) throw new StreamError("SocketStream: unable to create socket");

		if(connect(skt, &addr, addr.size))
			throw new StreamError("SocketStream: unable to establish connection");
	}

	void listenTo(ushort port)
	{
		sockaddr_in addr;
		addr.sin_family = 2;
		addr.sin_port = port;
		addr.sin_addr.s_addr = 0x0100007F; // localhost = 127.0.0.1

		int lst = socket(addr.sin_family, 1, 0);
		if(lst == -1) throw new StreamError("SocketStream: unable to create listening socket");

		if(bind(lst, &addr, addr.size))
			throw new StreamError("SocketStream: unable to bind socket");

		if(listen(lst, 2))
			throw new StreamError("SocketStream: unable to listen to port");

		uint addrsz = addr.size;
		skt = accept(lst, &addr, &addrsz);
		if(skt == -1) throw new StreamError("SocketStream: unable to create socket");

	    version(Win32)
	    {
		closesocket(lst);
	    }
	    version(linux)
	    {
		close(lst);
	    }
	}

	void disconnect()
	{
	    version(Win32)
	    {
		closesocket(skt);
		skt = -1;
	    }
	    version(linux)
	    {
		close(skt);
		skt = -1;
	    }
	}

    version(Win32)
    {
	static this()
	{
		WSAData data;
		if(WSAStartup(0x0101, &data))
			throw new StreamError("SocketStream: unable to initialize driver");
	}
	static ~this()
	{
		WSACleanup();
	}
    }

	// get(), seek(), position(), length()
	// and truncate() can not be implemented!

	ubyte[] get()
	{
		throw new StreamError("SocketStream: get() function is not supported");
		return null;
	}

	void seek(int offset, Rel rel)
	{
		throw new SeekError("SocketStream: seek() function is not supported");
	}

	ulong position()
	{
		throw new StreamError("SocketStream: position() function is not suppoted");
		return 0;
	}

	ulong length()
	{
		throw new StreamError("SocketStream: length() function is not supported");
		return 0;
	}

	void truncate()
	{
		throw new StreamError("SocketStream: truncate() function is not supported");
	}

		protected:

	int skt = -1;

	void readBlock(void* buf, uint buflen)
	{
		char* begin = cast(char*) buf;
		char* end = begin + buflen;

		while(begin < end)
		{
			int recdb = recv(skt, begin, end - begin, 0);
			if(recdb ==  0) throw new ReadError("SocketStream: connection closed");
			if(recdb == -1) throw new ReadError("SocketStream: unable to receive data");
			begin += recdb;
		}
	}

	void writeBlock(void* buf, uint buflen)
	{
		char* begin = cast(char*) buf;
		char* end = begin + buflen;

		while(begin < end)
		{
			int sentb = send(skt, begin, end - begin, 0);
			if(sentb == -1) throw new WriteError("SocketStream: unable to send data");
			begin += sentb;
		}
	}
}